package gs.blob.v0

import gs.hex.v0.Hex
import gs.hex.v0.HexDecode
import gs.hex.v0.HexEncode
import java.util.Base64

/** Opaque type that represents an array of bytes -- this type is intended to be
  * used for arbitrary blobs of data.
  */
opaque type Blob = Array[Byte]

object Blob:

  /** Instantiate a new [[Blob]].
    *
    * @param data
    *   The raw data.
    * @return
    *   The new blob.
    */
  def apply(data: Array[Byte]): Blob = data

  given CanEqual[Blob, Blob] = CanEqual.derived

  given HexEncode[Blob] = new HexEncode[Blob] {

    /** @inheritDocs
      */
    override def toHexString(data: Blob): String = Hex.toHexString(data)
  }

  given HexDecode[Blob] = new HexDecode[Array[Byte]] {

    /** @inheritDocs
      */
    override def fromHexString(data: String): Option[Array[Byte]] =
      Hex.fromHexString(data)

  }

  /** Convert the given base64 string to a [[Blob]].
    *
    * @param data
    *   The candidate data.
    * @return
    *   The parsed [[Blob]], or `None` if the input is not valid base64.
    */
  def fromBase64(data: String): Option[Blob] =
    scala.util.Try(Base64.getDecoder().decode(data)).toOption

  /** Convert the given hex string to a [[Blob]].
    *
    * @param data
    *   The candidate data.
    * @return
    *   The parsed [[Blob]], or `None` if the input is not valid base64.
    */
  def fromHex(data: String): Option[Blob] =
    HexDecode[Blob].fromHexString(data)

  extension (blob: Blob)
    /** @return
      *   The underlying bytes of this [[Blob]].
      */
    def toByteArray(): Array[Byte] = blob

    /** @return
      *   Hexadecimal string representation of this [[Blob]].
      */
    def toHex(): String = HexEncode[Blob].toHexString(blob)

    /** @return
      *   The base64 representation of this [[Blob]].
      */
    def toBase64(): String = Base64.getEncoder().encodeToString(blob)

    /** @return
      *   The number of bytes contained in this [[Blob]].
      */
    def length(): Int = blob.length

    /** Compare this blob to another in O(N) time. Determines if the bytes are
      * identical.
      *
      * @param other
      *   The blob to compare against.
      * @return
      *   True if the blobs are identical, false otherwise.
      */
    def sameBytesAs(other: Blob): Boolean = blob.sameElements(other)

end Blob
